/*
 *
 * Component:
 *
 * Description:
 *
 *
 * Last Updated     : $Author$
 * Update Date Time : $Date$
 * RCS File Name    : $RCSfile$
 * RCS Revision     : $Revision$
 * Locked by        : $Locker$
 * Symbolic name    : $Name$
 * State            : $State$
 */

#include <dcmtk/ofstd/ofstring.h>

/**
 * Holds an Original - New UID pair. Original UIDs will be replaced to New UIDs in DICOM objects
 */
class UidPair {
private:
	
	/** Original UID */
	OFString m_origUID;

	/** New UID */
	OFString m_newUID;

public:

	/**
	 * Constructor for creating a new UidPair object
	 *
	 * @param origUID  Original UID
	 * @param newUID   New UID
	 */
	UidPair( const OFString & origUID, const OFString & newUID ) {
		m_origUID = origUID;
		m_newUID = newUID;
	}

	/**
	 * Destructor for destroyin the UidPair object
	 */
	~UidPair() {
	}

	/**
	 * Checks if the Original UID of this UidPair object is the same as the argument
	 *
	 * @param uid UID to be checked
	 *
	 * @return true if and only if the Original UID of this UidPair object is the same as the argument
	 */
	inline bool equals( const OFString & uid ) const {
		return uid.compare( m_origUID ) == 0;
	}

	/**
	 * Retireves the Original UID of this UidPair object
	 *
	 * @return the Original UID
	 */
	inline const OFString & getOrigUid() const {
		return m_origUID;
	}

	/**
	 * Retireves the New UID of this UidPair object
	 *
	 * @return the New UID
	 */
	inline const OFString & getNewUid() {
		return m_newUID;
	}

	/**
	 * Copy constructor
	 *
	 * param origUidPair  Original UidPair object to be copied
	 */
	UidPair( const UidPair & origUidPair ) {
		m_origUID = origUidPair.m_origUID;
		m_newUID  = origUidPair.m_newUID;
	}

	/**
	 * Assignment operator
	 *
	 * param origUidPair  copy values into this UidPair object
	 *
	 * @return reference to this
	 */
	UidPair & operator=( const UidPair & origUidPair ) {
		m_origUID = origUidPair.m_origUID;
		m_newUID  = origUidPair.m_newUID;
		return *this;
	}
};


/**
 * Holds a list of UidPairs with UIDs to be replaced
 */
class UidList {
private:

	/** List of UidPairs */
	OFList<UidPair> m_list;

public:

	/**
	 * Constructor for creating a new UidList object
	 */
	UidList() {
	}

	/**
	 * Destructor for destroying this UidList object
	 */
	~UidList() {
		m_list.clear();
	}

	/**
	 * Retrieves New UID for an Original UID
	 *
	 * @param uid   Original UID to be searched for
	 * @param res   output parameter for holding New UID (if Original UID has been found)
	 *
	 * @return true if and only if Original UID has been found
	 */
	bool getPairFor(const OFString & uid, OFString & res);

	/**
	 * Adds an Origianal UID - New UID pair to the list
	 *
	 * @param origUID  Original UID
	 * @param newUID   New UID
	 */
	void addUidPair( const OFString & origUid, const OFString & newUid );

	/**
	 * Retrieves New UID for an Original UID, or generate a New UID for the
	 * Original UID if there was no New UID generated yet
	 *
	 * @param origUID  Original UID
	 * @param newUID   New UID
	 * @param uidroot  UID root to be used when generating new UID
	 */
	void getOrGenUidPair( const OFString & origUid, OFString & newUid, const char * uidroot );

	/**
	* Prints the content of the list to a file descriptor
	*
	* @param f filedescriptor to print on
	*/
	void print( FILE * f ) const;

	/**
	 * Loads Original UID - New UID pairs from a file.
	 * 
	 * @param fileName   Full path to the file to read
	 *
	 * @return true if and only if UID pairs have been successfully read from the file
	 */
	bool load( const OFString & fileName );
};



class CopyMaker {
private:

	/** List of Original UID - New UID pairs */
	UidList & m_list;
	
	/** Changes to be applied to DICOM objects during copy (used by merge/split) */
	DcmItem * m_modItem;
	
	/** PbR Dataset to get and apply modifications from Study Sequence (used by copy/forward) */
	DcmDataset * m_pbrDataset;
	
	/** Destination directory for object files to be copied to */
	OFString m_dstDir;

	/** Study Instance UID for destination study */
	OFString m_study;
	
	/** Should the saved files be registered automatically */
	bool m_needRegister;

	/** Counts the objects copied to new study */
	int m_copyCounter;

	/** Number of first N objects being registered directly (O: none, -1: all) */
	int m_registerDirectly;

	/** Job ID to be used to group the generated tasks to the same job */
	OFString m_jobId;

	/**
	 * Replaces UID references in PBFigure element
	 *
	 * @param elem   replace UID references in this PBFigure element
	 *
	 * @return true if and only if replacing UID references was successfull
	 */
	bool checkAndModifyFigure( DcmElement & elem );

	/**
	 * Replaces UID references in SoundReference element
	 *
	 * @param elem   replace UID references in this SoundReference element
	 *
	 * @return true if and only if replacing UID references was successfull
	 */
	bool checkAndModifySoundReference(DcmElement & elem);

	/**
	 * Replaces value in an UniqueIdentifier (UID) element
	 *
	 * @param elem   replace value of this UniqueIdentifier (UID) element
	 *
	 * @return true if and only if replacing value was successfull
	 */
	bool checkAndModifyUID( DcmElement & elem, const char * uidroot );
	
	/**
	 * Iterate through elements in item, and replace UID elements and UID references with new values
	 * The following elements will be modified:
	 *    SOP Instance UID
	 *    Series Instance UID
	 *    Study Instance UID
	 *    PB Referenced Results UID
	 *    PB Referenced Dictation UID
	 *    PB Referenced Interpretation UID
	 *    Referenced SOP Instance UID
	 *    PB Figure
	 *    Reference To Recorded Sound
	 *    Media Storage SOP Instance UID
	 *
	 * Referenced SOP Instance UID will only be replaced inside a Predecessor Documents Sequence
	 *
	 * Values for several elements can be specified in modItem.
	 * The values for these elements in item will be replaced from modItem.
	 *
	 * @param item      modify elements in this item
	 *
	 * @return true if and only if replacing UIDs and UID references was successfull
	 */
	bool modifyItem( DcmItem & item );

	void init( DcmItem * modItem, DcmDataset * pbrDataset, const OFString & dstDir ) {
		m_modItem = modItem;
		m_pbrDataset = pbrDataset;

		m_dstDir = dstDir;
		if ( m_dstDir[ m_dstDir.length() -1 ] != '/' )
			m_dstDir += "/";

		m_needRegister = false;
		m_copyCounter = 0;
		m_registerDirectly = 0;

		m_jobId = "";
	}
		
public:

	CopyMaker( UidList & uidList, DcmItem * modItem, DcmDataset * pbrDataset, const OFString & dstDir ) : m_list( uidList ) {
		init( modItem, pbrDataset, dstDir );
	}
		
	CopyMaker( UidList & uidList, const OFString & dstDir, const OFString & study) : m_list( uidList ) {
		init( NULL, NULL, dstDir );
		m_study = study;
	}

	~CopyMaker() {
	}


	/**
	 * Sets save mode after copying files.
	 * In default, objects will not be registered automatically to the database after save.
	 *
	 * @param needRegister    start registering saved objects to the database
	 * @param countDirect     number of N first object to be registered directly (0: none; -1: all) 
	 * 
	 */
	inline void setSaveMode( bool needRegister, int countDirect ) {
		m_needRegister = needRegister;
		m_registerDirectly = countDirect;
	}

	/**
	 * Assign PbR dataset having edit changes to be applied to objects being copied.
	 * Intended for Create Copy - for it needs to edit the PbR first.
	 *
	 * @param pbrDataset  PbR dataset having edit changes to be applied to objects being copied 
	 */
	inline void setPbRDataset( DcmDataset * pbrDataset ) {
		m_pbrDataset = pbrDataset;
	}
	
	/**
	 * Save DICOM object to destination directory.
	 * Intended for saving PbR object when creating copy
	 * 
	 * @param dcmff  DIOCM object to be saved
	 *
	 * @return true if and only if the DICOM object was successfully saved
	 */
	bool save( DcmFileFormat & dcmff );
	
	/**
	 * Apply UID changes and modifications to a single DICOM object.
	 *
	 * This method is intended to be used for forwarding shortcuts.
	 *
	 * @param dcmff  DIOCM object the changes to be applied to
	 *
	 * @return true if and only if changes have been successfully copied
	 */
	bool applyTo( DcmFileFormat & dcmff );

	/**
	 * Copying a single DICOM object by applying UID changes and modifications.
	 *
	 * @param src  source file name
	 *
	 * @return true if and only if the DICOM object has been successfully copied
	 */
	bool copyFile( const OFString & src );
};



